unit umain;

interface

uses Windows, Sysutils, Classes;

procedure main;

implementation

type
  PDWordArray = ^TDWordArray;
  TDWordArray = array[0..$0FFFFFFF] of DWord;

  TPartition = record
    Stream: TFileStream;
    PartType: Byte;
    PartBegin: DWord;
    BytsPerSec: Word;
    SecPerClus: Byte;
    BytsPerClus: Word;
    RsvdSecCnt: Word;
    NumFATs: Byte;
    FATSz32: Word;
    ClusCount: DWord;
    ClusBegin: DWord;
    RootClus: DWord;
    FAT: PDWordArray;
  end; 

function Min(i,j: Integer): Integer;
begin
  if i < j then
    Result:= i
  else
    Result:= j;
end;

// Muestra el contenido del buffer en hexadecimal y como texto
procedure WriteHex(Buffer: PAnsiChar;  Count: Integer);
var
  i,j: Integer;
begin
  j:= 0;
  while Count > 0 do
  begin
    Write(IntToHex(j,8) + ':' + #32#32);
    for i:= 0 to Min(Count,8) - 1 do
      Write(IntToHex(Byte(Buffer[i]),2) + #32);
    Write(#32);
    for i:= 8 to Min(Count,16) - 1 do
      Write(IntToHex(Byte(Buffer[i]),2) + #32);
    for i:= Min(Count,16) to 15 do
      Write(#32#32#32);
    Write(#32 + '|');
    for i:= 0 to Min(Count,16) - 1 do
    if Char(Buffer[i]) in ['A'..'Z','a'..'z','0'..'9'] then
      Write(Buffer[i])
    else
      Write('.');
    Writeln('|');
    Dec(Count,16);
    inc(Buffer,16);
    inc(j,16);
  end;
end;

function FormatFileSize(const Bytes: int64): String;
const
  B = 1;          // Byte
  KB = 1024 * B;  // Kilobyte
  MB = 1024 * KB; // Megabyte
  GB = 1024 * MB; // Gigabyte
begin
  if Bytes > GB then
    Result := FormatFloat('0.00 GB', Bytes / GB)
  else if Bytes > MB then
    Result := FormatFloat('0.00 MB', Bytes / MB)
  else if Bytes > KB then
    Result := FormatFloat('0.00 KB', Bytes / KB)
  else
    Result := FormatFloat(   '0   ', Bytes);
end;

// Lee el MasterBootRecord
procedure ReadMasterBootRecord(var Partition: TPartition);
var
  i,j: Integer;
  Buffer: PByteArray;
begin
  // Reservamos memoria para guardar el MBR
  GetMem(Buffer,$0200);
  try
    with Partition do
    begin
      // Leemos el MBR
      Stream.ReadBuffer(Buffer^,$0200);
      // Volcamos en pantalla el contenido del MBR
      Writeln('=== Master Boot Record ===');
      WriteHex(PAnsiChar(Buffer), $0200);
      Writeln;
      // Comprobamos la firma del MBR
      if PWORD(@Buffer[$01FE])^ <> $AA55 then
        raise Exception.Create('Invalid MBR signature = '
          + IntToHex(Swap(PWORD(@Buffer[$01FE])^),2));
      // Volcamos la tabla de particiones
      Writeln('=== Table ===');
      WriteHex(PAnsiChar(@Buffer[$01BE]), $0040);
      Writeln;
      // Volcamos cada entrada de la tabla de particiones
      Writeln('=== Partitions ===');
      for i:= 0 to 3 do
      begin
        j:= $01BE + (16 * i);
        if Buffer[j] = $80 then
          Writeln(Format('[%d] %8.8x/%8.8x A',[i,
            PDWORD(@Buffer[j + $08])^,PDWORD(@Buffer[j + $0C])^]))
        else
          Writeln(Format('[%d] %8.8x/%8.8x',[i,
            PDWORD(@Buffer[j + $08])^,PDWORD(@Buffer[j + $0C])^]))
      end;
      Writeln;
      // Usamos siempre la primera particion
      Writeln('Go to partition 0 ...');
      Writeln;
      PartType:= Buffer[$01BE + $04];
      PartBegin:= PDWORD(@Buffer[$01BE + $08])^;
    end;
  finally
    FreeMem(Buffer);
  end;
end;

// Lee el primer sector de la particion
procedure ReadBootSector(var Partition: TPartition);
var
  Buffer: PByteArray;
begin
  // Reservamos memoria para guardar el sector de arranque
  GetMem(Buffer,$0200);
  try
    with Partition do
    begin
      // Leemos el sector de arranque
      Stream.ReadBuffer(Buffer^,$0200);
      // Volcamos en pantalla el contenido del sector de arranque
      Writeln('=== Boot Sector ===');
      WriteHex(PAnsiChar(Buffer), $0200);
      Writeln;
      // Comprobamos la firma del sector de arranque
      if PWORD(@Buffer[$01FE])^ <> $AA55 then
        raise Exception.Create('Invalid Boot Sector signature = '
          + IntToHex(Swap(PWORD(@Buffer[$01FE])^),2));
      // Comprobamos que es FAT32
      if PartType <> 0 then
      begin
        // si no lo es mostramos el error
        if not (PartType in [$0B,$0C]) then
          raise Exception.Create('Invalid Partition Type = 0x'
            + IntToHex(PartType,2));
        Writeln('Partiton type:       0x' + IntToHex(PartType,2) + ' (FAT32)');
      end else
      begin
        // Si no queda mas remedio usamos el "System ID"
        if Trim(Copy(PAnsiChar(@Buffer[$0052]),1,8)) <> 'FAT32' then
          raise Exception.Create('Invalid System ID = '
            + Trim(Copy(AnsiString(PAnsiChar(@Buffer[$0052])),1,8)));
      end;
      // Mostramos la informacion de la particion por pantalla
      Writeln('Signature:           ' + IntToHex(Swap(PWORD(@Buffer[$01FE])^),2));
      Writeln('System ID:           ' + Trim(Copy(PAnsiChar(@Buffer[$0052]),1,8)));
      Writeln('OEM ID:              ' + Trim(Copy(PAnsiChar(@Buffer[$0003]),1,8)));
      Writeln('Volume label:        ' + Trim(Copy(PAnsiChar(@Buffer[$0047]),1,11)));
      Writeln('Volume serial:       ' + IntToHex(PDWORD(@Buffer[$0043])^,8));
      BytsPerSec:= PWORD(@Buffer[$000B])^;
      Writeln('Bytes per sector:    ' + IntToStr(BytsPerSec));
      SecPerClus:= PByte(@Buffer[$000D])^;
      Writeln('Sectors per cluster: ' + IntToStr(SecPerClus));
      BytsPerClus:= BytsPerSec * SecPerClus;
      Writeln('Bytes per cluster:   ' + IntToStr(BytsPerClus));
      RsvdSecCnt:= PWORD(@Buffer[$000E])^;
      Writeln('Reserved sectors:    ' + IntToStr(RsvdSecCnt));
      NumFATs:= PByte(@Buffer[$0010])^;
      Writeln('Number of FATs:      ' + IntToStr(NumFATs));
      FATSz32:= PWORD(@Buffer[$0024])^;
      Writeln('Sectors per FAT:     ' + IntToStr(FATSz32));
      ClusCount:= ((FATSz32 * BytsPerSec) div SizeOf(DWord));
      Writeln('Cluster count:       ' + IntToStr(ClusCount));
      ClusBegin:= RsvdSecCnt + (NumFATs * FATSz32);
      RootClus:= PDWORD(@Buffer[$002C])^;
      Writeln('Root cluster:        ' + IntToStr(RootClus));
      Writeln;
    end;
  finally
    // Liberamos la memoria
    FreeMem(Buffer);
  end;
end;

// Lee la FAT
procedure ReadFAT(Partition: TPartition);
var
  i: DWord;
  FreeClus: int64;
  ResClus: int64;
  BadClus: int64;
begin
  FreeClus:= 0;
  ResClus:= 0;
  BadClus:= 0;
  with Partition do
  begin
    // No colocamos al principio de la FAT
    Stream.Position:= (PartBegin + RsvdSecCnt) * BytsPerSec;
    // La cargamos en memoria
    Stream.ReadBuffer(FAT^, FATSz32 * BytsPerSec);
    Writeln('=== Clusters ===');
    // La recorremos registro a registro
    for i:= 0 to ClusCount - 1 do
    begin
      // y contamos el estado de cada cluster
      case FAT[i] of
        0: inc(FreeClus);
        $FFFFFF7: inc(BadClus);
      else
        inc(ResClus);
      end;
    end;
    // Mostramos los contadores
    Writeln('Free: ' + IntToStr(FreeClus));
    Writeln('Used: ' + IntToStr(ResClus));
    Writeln('Bad:  ' + IntToStr(BadClus));
    Writeln;
  end;
end;

// Escribe la FAT en el disco
procedure WriteFAT(Partition: TPartition);
var
  i: Byte;
begin
  with Partition do
  begin
    Stream.Position:= (PartBegin + RsvdSecCnt) * BytsPerSec;
    // La escribimos tantas veces como indique el parametro NumFATs
    for i:= 1 to NumFATs do
      Stream.WriteBuffer(FAT^, FATSz32 * BytsPerSec);
  end;
end;

// Lee un cluster
procedure ReadClus(Partition: TPartition; Cluster: DWord; Buffer: PByte);
begin
  with Partition do
  begin
    Stream.Position:= (PartBegin + ClusBegin + (Cluster - 2) * SecPerClus)
      * BytsPerSec;
    Stream.ReadBuffer(Buffer^, BytsPerClus);
  end;
end;

// Escribe un cluster
procedure WriteClus(Partition: TPartition; Cluster: DWord; Buffer: PByte);
begin
  with Partition do
  begin
    Stream.Position:= (PartBegin + ClusBegin + (Cluster - 2) * SecPerClus)
      * BytsPerSec;
    Stream.WriteBuffer(Buffer^, BytsPerClus);
  end;
end;

// Lee los datos de un fichero a partir de su primer cluster
procedure ReadFileData(Partition: TPartition; Cluster: DWord; Stream: TStream);
var
  Buffer: PByte;
begin
  // Reservamos memoria para guardar los datos de un cluster
  GetMem(Buffer,Partition.SecPerClus * Partition.BytsPerSec);
  try
    // Mientras no se llegue al final del fichero
    while Cluster < Partition.ClusCount do
    begin
      // Leemos el cluster
      ReadClus(Partition, Cluster, Buffer);
      // Lo grabamos en el stream
      Stream.WriteBuffer(Buffer^, Partition.SecPerClus * Partition.BytsPerSec);
      // Buscamos en la FAT el siguiente cluster
      Cluster:= Partition.FAT[Cluster];
    end;
  finally
    FreeMem(Buffer);
  end;
end;

// Muestra la estructura de un directorio
procedure Tree(Partition: TPartition; Cluster: DWord; Path: String);
var
  Tmp: TMemoryStream;
  Entry: PAnsiChar;
  Name: String;
  Attr: Byte;
  Size: DWord;
  FTime: TFILETIME;
  STime: TSYSTEMTIME;
  FTimeStr: String;
begin
  // Creamos un stream temporal
  Tmp:= TMemoryStream.Create;
  // Reservamos memoria para guardar cada entrada del directorio
  GetMem(Entry, 32);
  try
    // Leemos todas las entradas del directorio
    ReadFileData(Partition,Cluster,Tmp);
    Tmp.Position:= 0;
    // Procesamos cada una de las entradas del directorio
    while Tmp.Read(Entry^, 32) = 32 do
    begin
      // Si el primer byte es cero, significa que es la ultima
      if Entry[0] = #0 then
        break;
      // Si el primer bytes es alguno de estos, ignoramos la entrada
      if Entry[0] in [#$00,#$05,#$2E,#$E5] then
        continue;
      Attr:= Byte(Entry[$0B]);
      // Ignoramos los nombres largos
      if (Attr and $0F) = $0F then
        continue;
      // Obtenemos la extension
      Name:= Trim(Copy(Entry,9,3));
      if Name <> EmptyStr then
        Name:= '.' + Name;
      // y se la sumamos al nombre
      Name:= Trim(Copy(Entry,1,8)) + Name;
      // Obtenemos el tamao
      Size:= PDWord(@Entry[$1C])^;
      // La fecha de modificacin
      DosDateTimeToFileTime(PDWord(@Entry[$18])^,PDWord(@Entry[$16])^,FTime);
      FileTimeToSystemTime(Ftime, STime);
      FTimeStr:= FormatDateTime('dd/mm/yyyy  hh:nn', SystemTimeToDateTime(STime));
      // Si es un directorio
      if Attr and faDirectory = faDirectory then
      begin
        Writeln(Format('%s  %10.10s  %s%s',[FTimeStr,'<DIR>',Path,Name]));
        // volvemos a llamar esta funcion de forma recursiva
        Tree(Partition, (PWord(@Entry[$14])^ * $10000) + PWord(@Entry[$1A])^,
          Path + Name + '\');
      end else
        Writeln(Format('%s  %10.10s  %s%s',[FTimeStr,FormatFileSize(Size),Path,Name]));
    end;
  finally
    // Liberamos la memoria y el stream
    FreeMem(Entry);
    Tmp.Free;
  end;
end;

// Esta es la funcion principal
procedure main;
var
  C: Char;
  i: Integer;
  H: Integer;
  Str: String;
  LeerMBR: Boolean;
  Stream: TFileStream;
  Partition: TPartition;
begin
  try
    // Si el archivo como parametro
    if (ParamCount > 0) and (Copy(ParamStr(1),1,1) <> '/') then
    begin
      LeerMBR:= FindCmdLineSwitch('MBR',TRUE);
      Stream:= TFileStream.Create(ParamStr(1),fmOpenRead or fmShareDenyWrite);
    end else
    begin
      with TStringList.Create do
      try
        Add(EmptyStr);
        Writeln('[' + IntToStr(Count-1) + '] Salir');
        // Buscamos todas las unidades
        for C:= 'A' to 'Z' do
        begin
          Str:= '\\.\' + C + ':';
          H:= FileOpen(PAnsiChar(Str),fmOpenRead or fmShareDenyNone);
          // Si se puede abrir lo aadimos a la lista
          if H > 0 then
          begin
            CloseHandle(H);
            AddObject(Str,TObject(0));
            Writeln('[' + IntToStr(Count-1) + '] ' + Str);
          end;
        end;
        // Buscamos todos los discos (al menos los 16 primeros)
        for i:= 0 to 16 do
        begin
          Str:= '\\.\PhysicalDrive' + IntTostr(i);
          H:= FileOpen(PAnsiChar(Str),fmOpenRead or fmShareDenyNone);
          // Si se puede abrir lo aadimos a la lista
          if H > 0 then
          begin
            CloseHandle(H);
            AddObject(Str,TObject(1));
            Writeln('[' + IntToStr(Count-1) + '] ' + Str);
          end;
        end;
        i:= -1;
        while i < 0  do
        begin
          Write('Escoge: ');
          Readln(Str);
          i:= StrToIntDef(Str,-1);
        end;
        Writeln;
        if i > 0 then
        begin
          Stream:= TFileStream.Create(Strings[i],fmOpenRead or fmShareDenyWrite);
          LeerMBR:= Integer(Objects[i]) = 1;
        end else
          Exit;
      finally
        Free;
      end;
    end;
    if Stream = nil then
      Halt;
    try
      // Borramos todas las variables
      Fillchar(Partition,Sizeof(Partition),#0);
      Partition.Stream:= Stream;
      // Si nos pasan el parametro /MBR
      if LeerMBR then
      begin
        // Leemos el MBR y buscamos la primera particion
        ReadMasterBootRecord(Partition);
        Stream.Position:= 512 * Partition.PartBegin;
      end;
      // Leemos el sector de arranque de la particion
      ReadBootSector(Partition);
      // Reservamos espacio en memoria para la FAT
      with Partition do
        GetMem(FAT,FATSz32 * BytsPerSec);
      try
        // Leemos la FAT y la guardamos en memoria
        ReadFAT(Partition);
        // Volcamos la informacion en pantalla
        Tree(Partition, Partition.RootClus, '\');
      finally
        // Liberamos la memoria reservada para la FAT
        FreeMem(Partition.FAT);
      end;
    finally
      // Liberamos el stream
      Stream.Free;
    end;
  except
    // Si ocurre un error
    On E: Exception do
    begin
      // Lo mostramos
      Writeln('Exception: ' + E.Message);
      Writeln('LastError: ' + SysErrorMessage(GetLastError));
      Writeln('ParamStr(1): ' + ParamStr(1));
    end;
  end;
  Writeln;
  Writeln('Pulsa [ENTER] para cerrar el programa ...');
  Readln;
end;

end.
